/*
Copyright 2019 The KubeOne Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package externalccm

import (
	"context"
	"time"

	"k8c.io/kubeone/pkg/fail"
	"k8c.io/kubeone/pkg/state"

	corev1 "k8s.io/api/core/v1"
	"k8s.io/apimachinery/pkg/util/wait"
)

const (
	uninitializedTaint = "node.cloudprovider.kubernetes.io/uninitialized"
)

// Ensure external CCM deployen if Provider.External
func Ensure(s *state.State) error {
	if !s.Cluster.CloudProvider.External {
		return nil
	}

	return waitForInitializedNodes(s)
}

func waitForInitializedNodes(s *state.State) error {
	s.Logger.Info("Waiting for nodes to initialize by CCM...")

	return wait.PollUntilContextTimeout(s.Context, 5*time.Second, 10*time.Minute, false, func(ctx context.Context) (bool, error) {
		nodes := corev1.NodeList{}

		if err := s.DynamicClient.List(ctx, &nodes); err != nil {
			return false, fail.KubeClient(err, "listing Nodes")
		}

		for _, node := range nodes.Items {
			for _, taint := range node.Spec.Taints {
				if taint.Key == uninitializedTaint && taint.Value == "true" {
					return false, nil
				}
			}
		}

		return true, nil
	})
}
